#include <gdk-pixbuf/gdk-pixbuf.h>
#include <librsvg/rsvg.h>
#include <SDL.h>
#include <caml/mlvalues.h>
#include <caml/callback.h>
#include <caml/fail.h>
#include <caml/memory.h>
#include <caml/alloc.h>

#include "sdlvideo_stub.h"

CAMLprim value g_type_init_stub(value unit)
{
  g_type_init();
  return Val_unit;
}

#ifndef NDEBUG
static void pixbuf_finalize(void *pixbuf)
{
  printf("hop, je drfrence\n");
  g_object_unref(G_OBJECT(pixbuf));
}
#endif

/* Create a Sdlvideo.surface from a GdkPixbuf.
   Assumes the GdkPixbuf uses RGBA encoding with 8 bits per component (it should be 
   the case for librsvg-rendered pixbufs).
   Pixel data is shared between the two structures.
   Garbage collecting of the surface will trigger the finalizer for Sdlvideo.surface,
   which will call SDL_FreeSurface and g_object_unref on the GdkPixbuf.
*/
static value sdl_surface_of_gdk_pixbuf(GdkPixbuf *pixbuf)
{
  SDL_Surface *surface;
  if(pixbuf == NULL)
    failwith("sdl_surface_of_gdk_pixbuf");
#ifndef NDEBUG
  fprintf(stderr, "pixbuf : w = %d, h = %d, rowstride = %d, n_channels = %d, bpp = %d\n",
	  gdk_pixbuf_get_width(pixbuf),
	  gdk_pixbuf_get_height(pixbuf),
	  gdk_pixbuf_get_rowstride(pixbuf),
	  gdk_pixbuf_get_n_channels(pixbuf),
	  gdk_pixbuf_get_bits_per_sample(pixbuf));
#endif
  surface = SDL_CreateRGBSurfaceFrom(gdk_pixbuf_get_pixels(pixbuf),
				     gdk_pixbuf_get_width(pixbuf),
				     gdk_pixbuf_get_height(pixbuf),
				     gdk_pixbuf_get_n_channels(pixbuf)*gdk_pixbuf_get_bits_per_sample(pixbuf),
				     gdk_pixbuf_get_rowstride(pixbuf),
				     0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000);
  return Val_SDLSurface(surface, 1, Val_unit, 
#ifndef NDEBUG
			pixbuf_finalize,
#else
			g_object_unref,
#endif
			G_OBJECT(pixbuf));
}

CAMLprim value rsvg_pixbuf_from_file_mlstub(value file_name)
{
  return sdl_surface_of_gdk_pixbuf(rsvg_pixbuf_from_file(String_val(file_name), NULL));
}

CAMLprim value rsvg_pixbuf_from_file_at_zoom_mlstub(value file_name, value x_zoom, value y_zoom)
{
  return sdl_surface_of_gdk_pixbuf(rsvg_pixbuf_from_file_at_zoom(String_val(file_name), 
								 Double_val(x_zoom), Double_val(y_zoom),
								 NULL));
}

CAMLprim value rsvg_pixbuf_from_file_at_size_mlstub(value file_name, value width, value height)
{
  return sdl_surface_of_gdk_pixbuf(rsvg_pixbuf_from_file_at_size(String_val(file_name), 
								 Int_val(width), Int_val(height),
								 NULL));
}

CAMLprim value rsvg_pixbuf_from_file_at_max_size_mlstub(value file_name, value max_width, value max_height)
{
  return sdl_surface_of_gdk_pixbuf(rsvg_pixbuf_from_file_at_max_size(String_val(file_name), 
								     Int_val(max_width), 
								     Int_val(max_height),
								     NULL));
}

CAMLprim value rsvg_pixbuf_from_file_at_zoom_with_max_mlstub(value file_name, value x_zoom, value y_zoom, value max_width, value max_height)
{
  return sdl_surface_of_gdk_pixbuf(rsvg_pixbuf_from_file_at_zoom_with_max(String_val(file_name), 
									  Double_val(x_zoom),
									  Double_val(y_zoom),
									  Int_val(max_width), 
									  Int_val(max_height),
									  NULL));
}

struct ml_rsvg_data {
  RsvgHandle *h;
  value cb;
};

static
void ml_rsvg_size_callback(gint *w, gint *h, gpointer user_data)
{
  struct ml_rsvg_data *d = user_data;
  value r;
  r = callback2(d->cb, Val_int(*w), Val_int(*h));
  *w = Int_val(Field(r, 0));
  *h = Int_val(Field(r, 1));
}

CAMLprim value rsvg_handle_new_stub(value cb)
{
  RsvgHandle *h = rsvg_handle_new();
  struct ml_rsvg_data *d;
  value v;
  if(! h)
    raise_out_of_memory();
  d=stat_alloc(sizeof(*d));
  d->h = h;
  d->cb= cb;
  register_global_root(&d->cb);
  rsvg_handle_set_size_callback(h, ml_rsvg_size_callback, d, NULL);
  v=alloc_small(1, Abstract_tag);
  Field(v, 0)=(value)d;
  return v;
}

#define RsvgHandle_val(h) (((struct ml_rsvg_data*)(Field((h), 0)))->h)

CAMLprim value rsvg_handle_free_stub(value h)
{
  struct ml_rsvg_data *d = (struct ml_rsvg_data*)(Field((h), 0));
  rsvg_handle_free(d->h);
  remove_global_root(&d->cb);
  stat_free(d);
  return Val_unit;
}

CAMLprim value rsvg_handle_close_stub(value h)
{
  if(rsvg_handle_close(RsvgHandle_val(h), NULL) != TRUE)
    failwith("rsvg_handle_close");
  return Val_unit;
}

static inline 
void check_substring(value s, value o, value l)
{
  if(Int_val(o) < 0 || Int_val(l) < 0 || 
     Int_val(o) + Int_val(l) > string_length(s))
    invalid_argument("bad substring");
}

CAMLprim value rsvg_handle_write_stub(value h, value s, value off, value len)
{
  check_substring(s, off, len);
  if(rsvg_handle_write(RsvgHandle_val(h), 
		       String_val(s)+Int_val(off), Int_val(len), NULL) != TRUE)
    failwith("rsvg_handle_write");
  return Val_unit;
}

CAMLprim value rsvg_handle_get_pixbuf_stub(value h)
{
  return sdl_surface_of_gdk_pixbuf(rsvg_handle_get_pixbuf(RsvgHandle_val(h)));
}

/* void rsvg_handle_init (RsvgHandle * handle); */
/* CAMLprim value rsvg_handle_init_stub(value h) */
/* { */
/*   rsvg_handle_init(RsvgHandle_val(h)); */
/*   return Val_unit; */
/* } */

/* value rsvg_handle_set_dpi_stub(value h, value dpi) */
/* { */
/*   rsvg_handle_set_dpi(RsvgHandle_val(h), Double_val(dpi)); */
/*   return Val_unit; */
/* } */

/* value rsvg_handle_set_default_dpi_stub(value dpi) */
/* { */
/*   rsvg_handle_set_default_dpi(Double_val(dpi)); */
/*   return Val_unit; */
/* } */
